/* HASHING TABLE */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "asm.h"

#define TABLE_SIZE (1<<13)  /* size of hash table (must be power of 2) */
#define MAGICKA_KONSTANTA  13  /* proof MJ's magic constant */

struct info{
int address;   /* label address/value */
char *pointer; /* pointer to a label name */
unsigned lineno;    /* defining line no */
};

struct table_type{                    /* HASHING TABLE */
unsigned char count;
struct info *element;
};

static struct info *last_ele;
static struct table_type entry[TABLE_SIZE];
static unsigned  counter;


unsigned table_entries(void)
{
   return  counter;
}


/* hashing function */

static unsigned
hashl(char *slovo, unsigned len)
{
unsigned b=0;

while(len--)
{
   b+=slovo[len];
   b*=MAGICKA_KONSTANTA;
}
return b&(TABLE_SIZE-1);
}


static unsigned
hash(char *slovo)
{
return hashl(slovo,strlen(slovo));
}


int  next_table_entry(char **txt, int *value, unsigned *lineno)
{
static unsigned  a=0, c=0;
while (c >= entry[a].count)
{  c=0;  if (++a == TABLE_SIZE)  return a=0; }
*txt=entry[a].element[c].pointer;
*value=entry[a].element[c].address;
*lineno=entry[a].element[c].lineno;
c++;
return 1;
}


/* adds item to hash table */
/* returns nullpointer in case of error else the table address of label */

int
add_to_table(char *txt,int value,unsigned lineno)
{
unsigned a,c;
a=hash(txt);
c=entry[a].count;
if (!(c&c-1))
{ entry[a].element=realloc(entry[a].element,(2*c+!c)*sizeof(struct info));
if (!entry[a].element)return 0; }
entry[a].element[c].pointer=txt;
entry[a].element[c].address=value;
entry[a].element[c].lineno=lineno;
last_ele= &entry[a].element[c];
counter++;
if (!++entry[a].count)
{fprintf(stderr,"Error: internal hash counter overflow\n"); return 0;}
return 1;
}


void update_last_added_entry(int value)
{
   last_ele->address=value;
}

/* tests if label is in table */
/* returns address of label if true, otherwise -1 */

int 
is_in_table(char *txt, unsigned len)
{
int a,b;

/* if not in table  we need address in first pass  to check JR distance */
/* but for EQU we need its true value for possible range check later */
a=len?hashl(txt,len):hash(txt);
if (!entry[a].count) return -1;
for (b=0;b<entry[a].count;b++)
   if (len && !strncmp(txt,entry[a].element[b].pointer,len))
      return entry[a].element[b].address;
   else if (!len && !strcmp(txt,entry[a].element[b].pointer))
      return entry[a].element[b].address;
return -1;
}


/* initializes hash table */

void 
hash_table_init(void)
{
int a;

for (a=0;a<TABLE_SIZE;a++)
 {
 entry[a].count=0;
 entry[a].element=NULL;
 }
counter=0;
}


/* removes hash table from memory */

void
free_hash_table(void)
{
int a,b;
for (a=0;a<TABLE_SIZE;a++)
 {
 for (b=0;b<entry[a].count;b++)
  free(entry[a].element[b].pointer);
 free(entry[a].element);
 }
}
